/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.fortiss.tooling.base.ui.editor.fx;

import static java.lang.Math.max;
import static org.fortiss.tooling.base.ui.editor.fx.ContextMenuUtil.createElementCompositionContext;
import static org.fortiss.tooling.base.ui.editor.fx.ContextMenuUtil.createPrototypeMenu;

import java.util.List;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerFeatures;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.change.Change;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IClickController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.base.ClickControllerBase;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.rectangular.RectangularResizableContentControllerBase;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.model.IModelChangeProvider;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.base.model.element.ElementPackage;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.kernel.extension.data.IElementCompositionContext;
import org.fortiss.tooling.kernel.service.IElementCompositorService;
import org.fortiss.tooling.kernel.ui.extension.IModelElementHandler;
import org.fortiss.tooling.kernel.ui.service.IModelEditorBindingService;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;

import javafx.scene.Node;
import javafx.scene.control.MenuItem;
import javafx.scene.input.MouseEvent;

/**
 * {@link RectangularResizableContentControllerBase} with a listener mechanism for the layout of the
 * connected {@link EObject model element}.
 * 
 * @author hoelzl
 */
public abstract class EObjectBasedRectangularResizableContentControllerBase
		extends RectangularResizableContentControllerBase {
	/** {@link IModelChangeProvider} for this controller */
	private final LayoutModelElementModelChangeProvider layoutModelChangeProvider;
	/** The click controller handling double-click to open editor. */
	private final IClickController openEditorDoubleClickController = new ClickControllerBase() {
		@Override
		public Change singleClick(MouseEvent event, Node node, DiagramCoordinate diagramLocation) {
			return EObjectBasedRectangularResizableContentControllerBase.super.getClickController(
					node, diagramLocation).singleClick(event, node, diagramLocation);
		}

		@Override
		public Change secondaryClick(MouseEvent event, Node node,
				DiagramCoordinate diagramLocation) {
			return EObjectBasedRectangularResizableContentControllerBase.super.getClickController(
					node, diagramLocation).secondaryClick(event, node, diagramLocation);
		}

		@Override
		public Change doubleClick(MouseEvent event, Node node, DiagramCoordinate diagramLocation) {
			EObject eo = getLayoutedME();
			IModelElementHandler<EObject> handler =
					IModelElementHandlerService.getInstance().getModelElementHandler(eo);
			if(handler != null) {
				eo = handler.handleOpenModelElementRequest(eo);
			}
			IModelEditorBindingService.getInstance().openInEditor(eo);
			return null; // no changes to model
		}
	};

	/** Constructor. */
	public EObjectBasedRectangularResizableContentControllerBase(IContentMVCBundle mvcb) {
		super(mvcb);
		if(!(mvcb.getModel() instanceof ILayoutedModelElement)) {
			throw new IllegalArgumentException(
					"Model element must be an instance of an ILayoutedModelElement.");
		}
		ILayoutedModelElement lme = (ILayoutedModelElement)mvcb.getModel();
		this.layoutModelChangeProvider = new LayoutModelElementModelChangeProvider(lme) {
			/** {@inheritDoc} */
			@Override
			protected boolean acceptNotification(Notification notification) {
				if(notification.getNotifier() != getLayoutedME()) {
					return false;
				}
				return isAnchorageFeature(notification.getFeature());
			}
		};
	}

	/** Returns the {@link ILayoutedModelElement} model element. */
	protected ILayoutedModelElement getLayoutedME() {
		// wild cast works: see constructor check with exception
		return (ILayoutedModelElement)getModel();
	}

	/** {@inheritDoc} */
	@Override
	public IModelChangeProvider getModelChangeProvider() {
		return layoutModelChangeProvider;
	}

	/** {@inheritDoc} */
	@Override
	public List<MenuItem> contextMenuContributions(Node node, DiagramCoordinate diagramLocation) {
		// outer-most grid areas of the diagram are inaccessible
		DiagramViewerFeatures features = getViewer().getFeatures();
		double x = max(features.getHorizontalSpacing(), diagramLocation.getX());
		double y = max(features.getVerticalSpacing(), diagramLocation.getY());

		// wild cast works: see constructor exception
		EObject modelParent = (EObject)getModel();
		IElementCompositionContext edc = createElementCompositionContext(modelParent, x, y, false,
				getViewer().getFeatures().getCurrentZoomFactor());
		return createPrototypeMenu(modelParent, edc);
	}

	/** {@inheritDoc} */
	@Override
	public void delete() {
		IElementCompositorService.getInstance().decompose(getLayoutedME());
	}

	/** Checks whether the given feature corresponds to an anchorage model element. */
	protected boolean isAnchorageFeature(Object feature) {
		return feature == ElementPackage.Literals.IHIERARCHIC_ELEMENT__CONNECTORS;
	}

	/** {@inheritDoc} */
	@Override
	public IClickController getClickController(Node node, DiagramCoordinate diagramLocation) {
		return openEditorDoubleClickController;
	}
}
