/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor.fx.visuals;

import static javafx.scene.paint.Color.BLACK;
import static javafx.scene.paint.Color.rgb;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.LINK_TARGET_ALLOWED_TAG;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.LINK_TARGET_DENIED_TAG;
import static org.fortiss.tooling.base.ui.editor.fx.visuals.CoordinateCorrections.ANCHOR_DIMENSION;
import static org.fortiss.tooling.base.ui.editor.fx.visuals.CoordinateCorrections.ANCHOR_INSET;
import static org.fortiss.tooling.base.ui.utils.LWFXEditorUtils.convertEOrientationToSide;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.getConnectorOffsetOrientation;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.model.layout.ILayout;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.model.layout.IOffsetLayout;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.model.layout.ISideLayout;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.elliptic.CircularContentAnchorageVisualBase;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.OffsetOrientation;

import javafx.geometry.Dimension2D;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.geometry.Side;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;

/**
 * {@link CircularContentAnchorageVisualBase} providing a default visual for ports attached to boxes
 * in standard boxes-and-wires diagrams based on {@link ILayoutedModelElement}s.
 * 
 * @author munaro
 */
public abstract class LayoutedCircularContentAnchorageVisualBase
		extends CircularContentAnchorageVisualBase implements ISideLayout, IOffsetLayout {

	/** Constructor. */
	public LayoutedCircularContentAnchorageVisualBase(IContentAnchorageMVCBundle mvcb) {
		super(mvcb);
		// TODO(#3877): Move type checks to a common base class.
		Object model = mvcb.getModel();
		if(!(model instanceof ILayoutedModelElement)) {
			throw new IllegalArgumentException(
					"Expected model of type ILayoutedModelElement, but was " +
							model.getClass().getSimpleName() + ".");
		}
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <T extends ILayout> T getLayout(Class<T> type) {
		return (T)this;
	}

	/** {@inheritDoc} */
	@Override
	public Rectangle2D getModelBounds() {
		Point2D locOnParent = getLocationOnParent();
		Dimension2D dim = getDimensions();
		return new Rectangle2D(locOnParent.getX(), locOnParent.getY(), dim.getWidth(),
				dim.getHeight());
	}

	/** {@inheritDoc} */
	@Override
	public Dimension2D getDimensions() {
		return ANCHOR_DIMENSION;
	}

	/** {@inheritDoc} */
	@Override
	protected double getInset() {
		return ANCHOR_INSET;
	}

	/** {@inheritDoc} */
	@Override
	public double getOffset() {
		OffsetOrientation offsetOrientation = getConnectorOffsetOrientation(getLayoutedModelElement());
		double inset = getDimensions().getWidth() / 2;
		return offsetOrientation.getOffset() - inset;
	}

	/** {@inheritDoc} */
	@Override
	public Side getSide() {
		// TODO (#3868): Remove null check.
		OffsetOrientation offsetOrientation = getConnectorOffsetOrientation(getLayoutedModelElement());
		EOrientation orientation = EOrientation.NORTH;
		if(offsetOrientation != null) {
			orientation = offsetOrientation.getOrientation();
		}
		return convertEOrientationToSide(orientation);
	}

	/** {@inheritDoc} */
	@Override
	protected Paint getBorderColor() {
		return BLACK;
	}

	/** {@inheritDoc} */
	@Override
	protected double getOpacity() {
		return 1.0;
	}

	/** {@inheritDoc} */
	@Override
	protected double getBorderWidth() {
		return 1.0;
	}

	/** {@inheritDoc} */
	@Override
	protected Color getInteractionColor() {
		IContentAnchorageMVCBundle mvcBundle = getMVCBundle();
		if(mvcBundle.hasTag(LINK_TARGET_DENIED_TAG)) {
			return rgb(255, 0, 0, .25);
		}
		if(mvcBundle.hasTag(LINK_TARGET_ALLOWED_TAG)) {
			return rgb(0, 255, 0, .25);
		}
		return super.getInteractionColor();
	}

	/** Return the {@link ILayoutedModelElement}. */
	// TODO(#3877): Implement an equivalent, but generic method in a common base class.
	private ILayoutedModelElement getLayoutedModelElement() {
		// Safe wild cast due to type check in constructor
		return (ILayoutedModelElement)getModel();
	}
}
