/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import static org.eclipse.jface.dialogs.MessageDialog.openInformation;
import static org.eclipse.ui.PlatformUI.getWorkbench;

import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.TextCellEditor;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.service.ICommandStackService;

/**
 * {@link EditingSupport} for name of {@link INamedElement}.
 * 
 * (Used in featuremodel)
 * 
 * @author bayha
 */
public abstract class AbstractNameEditingSupport extends EditingSupport {

	/** Constructor. */
	public AbstractNameEditingSupport(ColumnViewer viewer) {
		super(viewer);
	}

	/**
	 * For a given {@link INamedElement} 'element', retrieves all siblings, among which 'element' is
	 * supposed to have a unique name.
	 * 
	 * @param element
	 *            {@link INamedElement} which shall have a unique name.
	 * @return An array of {@link INamedElement}, which contains all siblings for the
	 *         unique name contraint.
	 */
	protected abstract INamedElement[] getSiblingsForUniqueNameCheck(INamedElement element);

	/** {@inheritDoc} */
	@Override
	protected void setValue(Object element, final Object value) {
		final INamedElement elt = (INamedElement)element;
		if(value != null && !(value.toString().equals(elt.getName()))) {
			for(INamedElement child : getSiblingsForUniqueNameCheck(elt)) {
				if(child != element && value.equals(child.getName())) {
					Shell shell = getWorkbench().getActiveWorkbenchWindow().getShell();
					openInformation(shell, "Information",
							"This name already exists, please choose another one.");
					return;
				}
			}
			ICommandStackService.getInstance().runAsCommand(elt, new Runnable() {
				@Override
				public void run() {
					setValueInternal(value, elt);
				}
			});
		}
	}

	/** Performs the actual setName(...) call. */
	protected void setValueInternal(final Object value, final INamedElement elt) {
		elt.setName(value.toString());
	}

	/** {@inheritDoc} */
	@Override
	protected Object getValue(Object element) {
		return ((INamedElement)element).getName();
	}

	/** {@inheritDoc} */
	@Override
	protected CellEditor getCellEditor(Object element) {
		return new TextCellEditor((Composite)getViewer().getControl());
	}

	/** {@inheritDoc} */
	@Override
	protected boolean canEdit(Object element) {
		return element instanceof INamedElement;
	}
}
