/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *     Alexander Diewald (fortiss GmbH) - backport to AF3 & Integration
 *                                        improvements
 *
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.control.treetableview;

import javafx.beans.property.ReadOnlyStringWrapper;
import javafx.scene.Node;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.TreeItem;
import javafx.scene.control.TreeTableCell;
import javafx.scene.control.TreeTableColumn;
import javafx.scene.control.TreeTableView;

/**
 * A JavaFX {@link TreeTableView} based on an {@link DynamicTreeContentProviderBase} and
 * {@link DynamicTreeTableUIProviderBase}. This class is intended to be used as a wrapper for
 * {@link TreeTableView} and its rather complex expert API.
 * <P>
 * This class currently supports only a static tree-table layout, i.e., there is no way to alter the
 * columns unless the developer changes the underlying {@link TreeTableView} directly. Note that if
 * a column is removed the index provided to the label provider might not be correct anymore.
 * <P>
 * Furthermore, the selection model is currently fixed to be row selection.
 */
public final class DynamicTreeTableViewer<T> extends DynamicTreeViewerBase<T> {
	/** The {@link TreeTableView} control to be managed. */
	private final TreeTableView<T> view;
	/** The UI provider of this tree-table. */
	private final DynamicTreeTableUIProviderBase uiProvider;

	/** Constructor. */
	public DynamicTreeTableViewer(TreeTableView<T> view, T root, boolean showRoot, int revealLevel,
			DynamicTreeContentProviderBase<T> contentProvider,
			DynamicTreeTableUIProviderBase uiProvider) {
		super(contentProvider);
		this.uiProvider = uiProvider;
		// construct view
		this.view = view;
		DynamicTreeItem<T> rootItem = new DynamicTreeItem<T>(root, this);
		view.setRoot(rootItem);
		view.setShowRoot(showRoot);
		rootItem.update();
		// expand to reveal (+1 if root node is not shown
		expandItem(rootItem, showRoot ? revealLevel : revealLevel + 1);
	}

	/** Constructor. */
	public DynamicTreeTableViewer(T root, boolean showRoot, int revealLevel,
			DynamicTreeContentProviderBase<T> contentProvider,
			DynamicTreeTableUIProviderBase uiProvider) {
		this(new TreeTableView<T>(), root, showRoot, revealLevel, contentProvider, uiProvider);
	}

	/** Updates the viewer content. */
	public void update() {
		// wild cast works: see constructor
		DynamicTreeItem<T> rootItem = (DynamicTreeItem<T>)view.getRoot();
		rootItem.update();
	}

	/** Expands items up to the given level. */
	private void expandItem(TreeItem<T> parentItem, int revealLevel) {
		if(revealLevel <= 0) {
			return;
		}
		parentItem.setExpanded(true);
		revealLevel--;
		for(TreeItem<T> child : parentItem.getChildren()) {
			expandItem(child, revealLevel);
		}
	}

	/** Adds a column to the table part of the view. */
	public void addColumn(String headerLabel, int prefWidth) {
		final int num = view.getColumns().size();
		TreeTableColumn<T, String> col = new TreeTableColumn<>(headerLabel);
		col.setPrefWidth(prefWidth);
		col.setCellValueFactory(param -> {
			Object data = param.getValue().getValue();
			return new ReadOnlyStringWrapper(uiProvider.getLabel(data, num));
		});
		col.setCellFactory(param -> {
			TreeTableCell<T, String> cell = new TreeTableCell<T, String>() {
				@Override
				protected void updateItem(String item, boolean empty) {
					super.updateItem(item, empty);
					ContextMenu menu = null;
					Node icon = null;
					if(!empty && item != null) {
						T data = this.getTreeTableRow().getItem();
						menu = uiProvider.createContextMenu(data, num);
						icon = uiProvider.getIconNode(data, num);
					}
					this.setContextMenu(menu);
					this.setGraphic(icon);
				}
			};
			cell.textProperty().bind(cell.itemProperty());
			return cell;
		});
		view.getColumns().add(col);
	}

	/** Returns the underlying {@link TreeTableView}. */
	public TreeTableView<T> getControl() {
		return view;
	}
}
