/*******************************************************************************
 * Copyright (c) 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef;

import static java.lang.Math.floor;
import static java.lang.Math.max;
import static java.util.Objects.requireNonNull;
import static javafx.scene.paint.Color.LIGHTGRAY;

import javafx.geometry.Dimension2D;
import javafx.geometry.Point2D;
import javafx.scene.paint.Color;

/**
 * This class handles the state of all features of the {@link DiagramViewer}, which can be toggled
 * after instantiating the diagram viewer.
 */
public final class DiagramViewerFeatures {
	/** The diagram viewer. */
	private final DiagramViewer viewer;
	/** The zoom factors. */
	private double[] zoomFactors = new double[] {0.5, 0.75, 1, 1.5, 2};
	/** The current zoom factor array index. */
	private int zoomFactorIndex = 2;
	/** Flag indicating whether link highlighting is enabled. */
	private boolean linkHighlightingEnabled = false;
	/** Flag if interaction area shading is active. */
	private boolean interactionAreaShadingEnabled = false;
	/** The horizontal and vertical spacing of the grid indicators. */
	private Dimension2D indicatorSpacing = new Dimension2D(20, 20);
	/** The flag indicating that this grid is used to snap elements when moving or resizing. */
	private boolean useSnapToGrid = true;
	/** The size of the grid indicators. */
	private double indicatorSize = 2;
	/** The color of the grid indicator. */
	private Color indicatorColor = LIGHTGRAY.darker();
	/** The flag for drawing the outer border, which occupies the first grid row and lane. */
	private boolean drawOuterBorder = true;
	/** The indicator type. */
	private IndicatorType indicatorType = IndicatorType.CROSS;
	/** The background color. */
	private Color backgroundColor = LIGHTGRAY;

	/** Constructor. */
	public DiagramViewerFeatures(DiagramViewer diagramViewer) {
		this.viewer = diagramViewer;
	}

	/** Returns the zoom factors. */
	public double[] getZoomFactors() {
		return zoomFactors;
	}

	/** Sets the zoom factors. */
	public void setZoomFactors(double[] zoomFactors) {
		this.zoomFactors = zoomFactors;
		if(zoomFactorIndex >= zoomFactors.length) {
			zoomFactorIndex = zoomFactors.length - 1;
		}
		viewer.updateAllVisuals();
	}

	/** Returns the zoom factor index. */
	public int getZoomFactorIndex() {
		return zoomFactorIndex;
	}

	/** Returns the current zoom factor. */
	public double getCurrentZoomFactor() {
		return zoomFactorIndex == -1 ? 1.0 : zoomFactors[zoomFactorIndex];
	}

	/** Sets the zoom factor index. */
	public void setZoomFactorIndex(int zoomFactorIndex) {
		this.zoomFactorIndex = zoomFactorIndex;
		viewer.updateAllVisuals();
	}

	/** Increases zoom factor index by one up to the maximum. */
	public void zoomIn() {
		if(zoomFactorIndex < zoomFactors.length - 1) {
			zoomFactorIndex++;
			viewer.updateAllVisuals();
		}
	}

	/** Decreases zoom factor index by one down to the minimum. */
	public void zoomOut() {
		if(zoomFactorIndex > 0) {
			zoomFactorIndex--;
			viewer.updateAllVisuals();
		}
	}

	/** Returns the link highlighting enabled flag. */
	public boolean isLinkHighlightingEnabled() {
		return linkHighlightingEnabled;
	}

	/** Sets the link highlighting enabled. */
	public void setLinkHighlightingEnabled(boolean linkHighlightingEnabled) {
		this.linkHighlightingEnabled = linkHighlightingEnabled;
		viewer.updateAllVisuals();
	}

	/** Returns the interaction area shading enabled flag. */
	public boolean isInteractionAreaShadingEnabled() {
		return interactionAreaShadingEnabled;
	}

	/** Sets the interaction area shading enabled flag. */
	public void setInteractionAreaShadingEnabled(boolean interactionAreaShadingEnabled) {
		this.interactionAreaShadingEnabled = interactionAreaShadingEnabled;
		viewer.updateAllVisuals();
	}

	/** Returns the indicator type. */
	public IndicatorType getIndicatorType() {
		return indicatorType;
	}

	/** Sets the indicator type. */
	public void setIndicatorType(IndicatorType indicatorType) {
		this.indicatorType = indicatorType;
		viewer.updateAllVisuals();
	}

	/** Returns the background color. */
	public Color getBackgroundColor() {
		return backgroundColor;
	}

	/** Sets the background color. */
	public void setBackgroundColor(Color backgroundColor) {
		this.backgroundColor = backgroundColor;
		viewer.updateAllVisuals();
	}

	/** Returns the indicator color. */
	public Color getIndicatorColor() {
		return indicatorColor;
	}

	/** Sets the indicator color. */
	public void setIndicatorColor(Color indicatorColor) {
		this.indicatorColor = indicatorColor;
		viewer.updateAllVisuals();
	}

	/** Returns the indicator size. */
	public double getIndicatorSize() {
		return indicatorSize;
	}

	/** Sets the indicator size. */
	public void setIndicatorSize(double indicatorSize) {
		this.indicatorSize = indicatorSize;
		viewer.updateAllVisuals();
	}

	/** Returns the indicator spacing. */
	public Dimension2D getIndicatorSpacing() {
		return indicatorSpacing;
	}

	/** Returns the horizontal spacing. */
	public double getHorizontalSpacing() {
		return indicatorSpacing.getWidth();
	}

	/** Returns the vertical spacing. */
	public double getVerticalSpacing() {
		return indicatorSpacing.getHeight();
	}

	/** Sets the indicator spacing. */
	public void setIndicatorSpacing(Dimension2D indicatorSpacing) {
		this.indicatorSpacing = requireNonNull(indicatorSpacing);
		viewer.updateAllVisuals();
	}

	/** Returns whether to use snap to grid. */
	public boolean useSnapToGrid() {
		return useSnapToGrid;
	}

	/** Sets the use snap to grid flag. */
	public void setUseSnapToGrid(boolean useSnapToGrid) {
		this.useSnapToGrid = useSnapToGrid;
		viewer.updateAllVisuals();
	}

	/** Computes the closest horizontal grid coordinate of the given point. */
	public double snapToGridX(double x) {
		if(useSnapToGrid) {
			double width = indicatorSpacing.getWidth();
			return floor(x / width) * width;
		}
		return x;
	}

	/** Computes the closest vertical grid coordinate of the given point. */
	public double snapToGridY(double y) {
		if(useSnapToGrid) {
			double height = indicatorSpacing.getHeight();
			return floor(y / height) * height;
		}
		return y;
	}

	/** Computes the closest grid location of the given point. */
	public Point2D snapToGrid(Point2D point) {
		return snapToGrid(point.getX(), point.getY());
	}

	/** Computes the closest grid location of the given point. */
	public Point2D snapToGrid(double x, double y) {
		return new Point2D(snapToGridX(x), snapToGridY(y));
	}

	/**
	 * Computes the closest horizontal coordinate of the given point that lies in the center between
	 * the grid points.
	 */
	public double snapToCenterX(double x) {
		double hs2 = indicatorSpacing.getWidth() / 2;
		return snapToGridX(x - hs2) + hs2;
	}

	/**
	 * Computes the closest vertical coordinate of the given point that lies in the center between
	 * the grid points.
	 */
	public double snapToCenterY(double y) {
		double vs2 = indicatorSpacing.getHeight() / 2;
		return snapToGridY(y - vs2) + vs2;
	}

	/** Computes the closest center location of the given point. */
	public DiagramCoordinate snapToCenter(DiagramCoordinate point) {
		return snapToCenter(point.getX(), point.getY());
	}

	/** Computes the closest center location of the given point. */
	public DiagramCoordinate snapToCenter(double x, double y) {
		return new DiagramCoordinate(snapToCenterX(x), snapToCenterY(y));
	}

	/** Returns the draw outer border flag. */
	public boolean isDrawOuterBorder() {
		return drawOuterBorder;
	}

	/** Sets the draw outer border. */
	public void setDrawOuterBorder(boolean drawOuterBorder) {
		this.drawOuterBorder = drawOuterBorder;
		viewer.updateAllVisuals();
	}

	/** The indicator type used for the grid points. */
	public static enum IndicatorType {
		/** Invisible indicator. */
		INVISIBLE,
		/** Circle indicator. */
		CIRCLE,
		/** Cross indicator. */
		CROSS,
		/** Square indicator. */
		SQUARE;
	}

	/** Returns the maximum of the horizontal and vertical spacing. */
	public double getMaximumSpacing() {
		return max(getVerticalSpacing(), getHorizontalSpacing());
	}
}
