/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef;

import static javafx.scene.input.MouseEvent.*;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.*;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.change.Change;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IClickController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IDragController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IKeyPressController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.IContentAnchorageVisual;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.IContentVisual;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.IVisual;

import javafx.event.EventHandler;
import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;

/**
 * Class for monitoring mouse operations including clicking, dragging and linking
 * gestures. The listeners are installed on every node that is added to the diagram layers.
 * This class stores a mapping from these nodes to {@link IMVCBundle}s in order to route
 * mouse gestures, clicks and key typing to the respective controller of the respective bundle.
 */
public final class MouseState {
	/** The diagram viewer. */
	private final DiagramViewer viewer;
	/** The inverse map for {@link IMVCBundle} lookup with nodes. */
	private final Map<Node, IMVCBundle> nodesToBundleMap = new HashMap<Node, IMVCBundle>();
	/**
	 * The last mouse down event, which is used for location correction during drag gestures. See
	 * in-code comment of {@link #mouseDragDetectedHandler}.
	 */
	private MouseEvent lastMousePressedEvent = null;
	/** Stores the last mouse location. */
	private DiagramCoordinate lastMouseLocation = new DiagramCoordinate(0, 0);
	/** Stores the drag start or {@code null} if there is no drag currently. */
	private DiagramCoordinate dragStartLocation = null;

	/** Constructor. */
	public MouseState(DiagramViewer viewer) {
		this.viewer = viewer;
	}

	/** The mouse clicked handler. */
	private EventHandler<MouseEvent> mouseClickedHandler = new EventHandler<MouseEvent>() {
		@Override
		public void handle(MouseEvent event) {
			if(!nodesToBundleMap.containsKey(event.getSource())) {
				return;
			}
			// wild cast works, since source is in nodesMap
			Node source = (Node)event.getSource();
			IMVCBundle bundle = nodesToBundleMap.get(source);
			lastMouseLocation =
					viewer.convertEventToDiagramCoordinates(event.getX(), event.getY(), source);
			IClickController cc =
					bundle.getController().getClickController(source, lastMouseLocation);
			if(cc != null) {
				Change chg = null;
				if(event.getButton() == MouseButton.SECONDARY) {
					if(event.getClickCount() == 1) {
						chg = cc.secondaryClick(event, source, lastMouseLocation);
					}
				} else if(event.getButton() == MouseButton.PRIMARY) {
					if(event.getClickCount() == 1) {
						chg = cc.singleClick(event, source, lastMouseLocation);
					} else if(event.getClickCount() == 2) {
						chg = cc.doubleClick(event, source, lastMouseLocation);
					}
				}
				event.consume();
				if(chg != null) {
					viewer.getModelModifier().accept(chg);
				}
			}
		}
	};

	/** The current drag controller is non-null while a drag gesture is in progress. */
	private IDragController dragController = null;

	/** The drag detected handler deciding about link drag and normal drag gestures. */
	private EventHandler<MouseEvent> mouseDragDetectedHandler = new EventHandler<MouseEvent>() {
		@Override
		public void handle(MouseEvent event) {
			if(!nodesToBundleMap.containsKey(event.getSource())) {
				return;
			}
			event.consume();
			// the drag detected event occurs a view pixels away from the mouse press location
			// (which is required to detect the drag gesture at all). However, we have to take
			// the mouse pressed location, since that location's cursor might imply a different
			// operation when the mouse is dragged the one or the other location (e.g. move cursor
			// is only a view pixels away from resize cursor). If we do not correct the location to
			// the initial mouse pressed location the drag controller initiates the wrong operation
			// (e.g. a resize operation although the user wanted a move operation).
			double x, y;
			Node source;
			if(lastMousePressedEvent != null) {
				x = lastMousePressedEvent.getX();
				y = lastMousePressedEvent.getY();
				source = (Node)lastMousePressedEvent.getSource();
			} else {
				x = event.getX();
				y = event.getY();
				source = (Node)event.getSource();
			}
			// wild cast works, since source is in nodesToBundleMap
			IMVCBundle bundle = nodesToBundleMap.get(source);
			IVisual visual = bundle.getVisual();
			lastMouseLocation = viewer.convertEventToDiagramCoordinates(x, y, source);
			EDragGesture edg = visual.getDragGesture(source, lastMouseLocation);
			if(edg == EDragGesture.NONE) {
				return;
			}
			dragController = viewer.createDragController(bundle, edg, source, lastMouseLocation);
			if(dragController != null) {
				dragController.dragStarted(bundle, source, lastMouseLocation);
				source.startFullDrag();
			}
		}
	};
	/** The drag completed handler for move, resize, and link gestures. */
	private EventHandler<MouseEvent> mouseDragCompletedHandler = new EventHandler<MouseEvent>() {
		@Override
		public void handle(MouseEvent event) {
			if(!nodesToBundleMap.containsKey(event.getSource())) {
				return;
			}
			if(dragController == null) {
				return;
			}
			event.consume();
			// wild cast works, since source is in nodesToBundleMap
			Node source = (Node)event.getSource();
			IMVCBundle bundle = nodesToBundleMap.get(source);
			lastMouseLocation =
					viewer.convertEventToDiagramCoordinates(event.getX(), event.getY(), source);
			Change chg = dragController.dragCompleted(bundle, source, lastMouseLocation);
			dragController = null;
			if(chg != null) {
				viewer.getModelModifier().accept(chg);
			}
		}
	};

	/** The mouse dragged handler. */
	private EventHandler<MouseEvent> mouseDraggedHandler = new EventHandler<MouseEvent>() {
		@Override
		public void handle(MouseEvent event) {
			if(!nodesToBundleMap.containsKey(event.getSource())) {
				return;
			}
			if(dragController == null) {
				// there are drag events even before the drag detected event fired
				// ignore those phantoms
				return;
			}
			event.consume();
			// wild cast works, since source is in nodesToBundleMap
			Node source = (Node)event.getSource();
			IMVCBundle bundle = nodesToBundleMap.get(source);
			lastMouseLocation =
					viewer.convertEventToDiagramCoordinates(event.getX(), event.getY(), source);
			dragController.dragInProgress(bundle, source, lastMouseLocation);
			viewer.updateAllVisuals();
		}
	};

	/** The mouse release handler. */
	private EventHandler<MouseEvent> mouseReleasedHandler = new EventHandler<MouseEvent>() {
		@Override
		public void handle(MouseEvent event) {
			if(lastMousePressedEvent != null &&
					event.getButton() == lastMousePressedEvent.getButton()) {
				lastMousePressedEvent = null;
			}
			if(!nodesToBundleMap.containsKey(event.getSource())) {
				return;
			}
			if(dragController == null) {
				return;
			}
			event.consume();
			// wild cast works, since source is in nodesToBundleMap
			Node source = (Node)event.getSource();
			IMVCBundle bundle = nodesToBundleMap.get(source);
			lastMouseLocation =
					viewer.convertEventToDiagramCoordinates(event.getX(), event.getY(), source);
			Change chg = dragController.dragCompleted(bundle, source, lastMouseLocation);
			dragController = null;
			if(chg != null) {
				viewer.getModelModifier().accept(chg);
			}
		}
	};

	/** The mouse enter/exit handler. */
	private EventHandler<MouseEvent> mouseEnterExitHandler = new EventHandler<MouseEvent>() {
		@Override
		public void handle(MouseEvent event) {
			if(!nodesToBundleMap.containsKey(event.getSource())) {
				return;
			}
			event.consume();
			// wild cast works, since source is in nodesMap
			Node source = (Node)event.getSource();
			IMVCBundle bundle = nodesToBundleMap.get(source);
			if(event.getEventType() == MOUSE_ENTERED) {
				bundle.addTag(HOVER_TAG);
				updateVisualsAfterHoverChange(bundle);
			} else if(event.getEventType() == MOUSE_EXITED) {
				bundle.removeTag(HOVER_TAG);
				updateVisualsAfterHoverChange(bundle);
			}
			// prevent cursor from changing during drag operations
			if(dragController != null) {
				return;
			}
			lastMouseLocation =
					viewer.convertEventToDiagramCoordinates(event.getX(), event.getY(), source);
			IController ctrl = bundle.getController();
			if(ctrl != null) {
				Cursor cursor = ctrl.getCurrentCursor(source, lastMouseLocation);
				source.setCursor(cursor);
			}
		}

		/** Updates visuals after the hover tag has changed. */
		private void updateVisualsAfterHoverChange(IMVCBundle bundle) {
			viewer.updateVisual(bundle);
			// the following two sections allow hidden anchorages, which are
			// only shown once the mouse hovers over their parent visual.
			if(bundle instanceof IContentMVCBundle) {
				IContentMVCBundle contentBundle = (IContentMVCBundle)bundle;
				IContentVisual contentVisual = contentBundle.getVisual();
				if(contentVisual.updateAttachedAnchorageVisualsOnHover()) {
					// update anchorage visuals
					contentBundle.getAnchorages().stream()
							.forEach(anchorageBundle -> viewer.updateVisual(anchorageBundle));
				}
			} else if(bundle instanceof IContentAnchorageMVCBundle) {
				IContentAnchorageMVCBundle anchorageBundle = (IContentAnchorageMVCBundle)bundle;
				IContentAnchorageVisual anchorageVisual = anchorageBundle.getVisual();
				if(anchorageVisual.updateSiblingAnchorageVisualsOnHover()) {
					IContentMVCBundle contentBundle = anchorageBundle.getAttachedTo();
					// update sibling visuals
					contentBundle.getAnchorages().stream().filter(ab -> ab != anchorageBundle)
							.forEach(ab -> viewer.updateVisual(ab));
				}
			}
		}
	};

	/** The mouse motion handler used for updating the cursor. */
	private EventHandler<MouseEvent> mouseMotionHandler = new EventHandler<MouseEvent>() {
		@Override
		public void handle(MouseEvent event) {
			if(!nodesToBundleMap.containsKey(event.getSource()) || dragController != null) {
				return;
			}
			event.consume();
			// wild cast works, since source is in nodesMap
			Node source = (Node)event.getSource();
			IMVCBundle bundle = nodesToBundleMap.get(source);
			IController ctrl = bundle.getController();
			lastMouseLocation =
					viewer.convertEventToDiagramCoordinates(event.getX(), event.getY(), source);
			if(ctrl != null) {
				Cursor cursor = ctrl.getCurrentCursor(source, lastMouseLocation);
				source.setCursor(cursor);
			}
		}
	};

	/** The mouse pressed handler for selecting the node under the cursor. */
	private EventHandler<MouseEvent> mousePressedHandler = new EventHandler<MouseEvent>() {
		@Override
		public void handle(MouseEvent event) {
			// see mouseDragDetectedHandler for explanation of storing the last mouse press event
			lastMousePressedEvent = event;
			viewer.hideContextMenu();
			event.consume();
		}
	};

	/** The key event handler. */
	private final EventHandler<KeyEvent> keyEventHandler = new EventHandler<KeyEvent>() {
		@Override
		public void handle(KeyEvent event) {
			if(!nodesToBundleMap.containsKey(event.getSource())) {
				return;
			}
			// wild cast works, since source is in nodesMap
			Node source = (Node)event.getSource();
			IMVCBundle bundle = nodesToBundleMap.get(source);
			IKeyPressController kpc =
					viewer.createKeyPressController(bundle, source, lastMouseLocation);
			if(kpc != null) {
				Change change = kpc.keyEvent(event, bundle, source, lastMouseLocation);
				if(change != null) {
					viewer.getModelModifier().accept(change);
				}
			}
		}
	};

	/** Registers the mouse listeners on the given node. */
	/* package */ void registerMouseListeners(Node node, IMVCBundle mvcb) {
		node.setOnMousePressed(mousePressedHandler);
		node.setOnMouseEntered(mouseEnterExitHandler);
		node.setOnMouseExited(mouseEnterExitHandler);
		node.setOnMouseClicked(mouseClickedHandler);
		node.setOnDragDetected(mouseDragDetectedHandler);
		node.setOnMouseDragged(mouseDraggedHandler);
		node.setOnMouseDragOver(mouseDraggedHandler);
		node.setOnMouseReleased(mouseReleasedHandler);
		node.setOnMouseDragReleased(mouseDragCompletedHandler);
		node.setOnMouseMoved(mouseMotionHandler);
		node.setOnKeyReleased(keyEventHandler);
		nodesToBundleMap.put(node, mvcb);
	}

	/** Unregisters the mouse listeners from the given node. */
	/* package */ void unregisterMouseListeners(Node node) {
		node.setOnKeyReleased(null);
		node.setOnMouseMoved(null);
		node.setOnMouseDragReleased(null);
		node.setOnMouseReleased(null);
		node.setOnMouseDragOver(null);
		node.setOnMouseDragged(null);
		node.setOnDragDetected(null);
		node.setOnMouseClicked(null);
		node.setOnMouseExited(null);
		node.setOnMouseEntered(null);
		node.setOnMousePressed(null);
		nodesToBundleMap.remove(node);
	}

	/** Returns the last known mouse location. */
	public DiagramCoordinate getLastMouseLocation() {
		return lastMouseLocation;
	}

	/** Returns the drag start location or {@code null} if no drag is in progress. */
	public DiagramCoordinate getDragStartLocation() {
		return dragStartLocation;
	}

	/**
	 * Returns the current drag extent, i.e., last mouse position minus drag start location, or
	 * {@code null} if no drag is in progress.
	 */
	public DiagramCoordinate getDragExtent() {
		if(dragStartLocation == null || lastMouseLocation == null) {
			return null;
		}
		return lastMouseLocation.subtract(dragStartLocation);
	}
}
