/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.impl;

import static java.util.Objects.requireNonNull;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewer;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.FeedbackChange;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.model.IModelChangeListener;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.model.IModelChangeProvider;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.MVCBundleTag;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.IVisual;

/**
 * This class bundles all objects used by the model-view-controller paradigm.
 * <p>
 * It is also responsible for propagating model changes to changes of the visual.
 * Model changes may be received directly from the model element (if it implements
 * {@link IModelChangeProvider}) or from the assigned controller (if that one
 * implements {@link IModelChangeProvider}).
 */
public abstract class MVCBundleBase implements IModelChangeListener, IMVCBundle {
	/** The model. */
	private final Object model;
	/** The diagram viewer to be used for selection handling. */
	private final DiagramViewer viewer;
	/** The visual. */
	private IVisual visual;
	/** The controller. */
	private IController controller;
	/** The current feedback change applied to this bundle's visual. */
	private FeedbackChange feedbackChange = new FeedbackChange();
	/** The set of {@link MVCBundleTag}s. */
	private final Map<String, MVCBundleTag> tagMap = new HashMap<>();

	/** Constructor. */
	public MVCBundleBase(Object m, DiagramViewer v) {
		model = requireNonNull(m, "Model must not be null.");
		viewer = requireNonNull(v, "Viewer must not be null.");
	}

	/** {@inheritDoc} */
	@Override
	public final void setVisual(IVisual visual) {
		this.visual = requireNonNull(visual, "Visual must not be null.");
	}

	/** {@inheritDoc} */
	@Override
	public void attach() {
		if(model instanceof IModelChangeProvider) {
			IModelChangeProvider mcp = (IModelChangeProvider)model;
			mcp.addListener(this);
		}
		if(controller instanceof IModelChangeProvider) {
			IModelChangeProvider mcp = (IModelChangeProvider)controller;
			mcp.addListener(this);
		}
		IModelChangeProvider adapted = controller.getModelChangeProvider();
		if(adapted != null) {
			adapted.addListener(this);
		}
	}

	/** {@inheritDoc} */
	@Override
	public final void detach() {
		if(model instanceof IModelChangeProvider) {
			IModelChangeProvider mcp = (IModelChangeProvider)model;
			mcp.removeListener(this);
		}
		if(controller instanceof IModelChangeProvider) {
			IModelChangeProvider mcp = (IModelChangeProvider)controller;
			mcp.removeListener(this);
		}
		IModelChangeProvider adapted = controller.getModelChangeProvider();
		if(adapted != null) {
			adapted.removeListener(this);
		}
	}

	/** {@inheritDoc} */
	@Override
	public final void setController(IController controller) {
		this.controller = requireNonNull(controller, "Controller must not be null.");
	}

	/** {@inheritDoc} */
	@Override
	public Object getModel() {
		return model;
	}

	/** {@inheritDoc} */
	@Override
	public IVisual getVisual() {
		return visual;
	}

	/** {@inheritDoc} */
	@Override
	public IController getController() {
		return controller;
	}

	/** {@inheritDoc} */
	@Override
	public DiagramViewer getViewer() {
		return viewer;
	}

	/** {@inheritDoc} */
	@Override
	public void changed(Object element) {
		viewer.updateFromModel();
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * Updates the visual after setting the feedback change.
	 */
	@Override
	public final void setFeedbackChange(FeedbackChange feedbackChange) {
		this.feedbackChange = requireNonNull(feedbackChange);
	}

	/** {@inheritDoc} */
	@Override
	public final FeedbackChange getFeedbackChange() {
		return feedbackChange;
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return getClass().getSimpleName() + " MVC bundle of " + getModel().toString();
	}

	/** {@inheritDoc} */
	@Override
	public void addTag(MVCBundleTag tag) {
		tagMap.put(tag.getTagIdentifier(), tag);
	}

	/** {@inheritDoc} */
	@Override
	public MVCBundleTag getTag(MVCBundleTag tag) {
		return tagMap.get(tag.getTagIdentifier());
	}

	/** {@inheritDoc} */
	@Override
	public boolean hasTag(MVCBundleTag tag) {
		if(tag == null) {
			return false;
		}
		return tagMap.get(tag.getTagIdentifier()) != null;
	}

	/** {@inheritDoc} */
	@Override
	public void removeTag(MVCBundleTag tag) {
		tagMap.remove(tag.getTagIdentifier());
	}
}
