/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.rectangular;

import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.rectangular.RectangularBorderLocation.getClosestLocationOnBounds;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramLayers;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.model.layout.IOffsetLayout;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.model.layout.ISideLayout;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.IContentAnchorageVisual;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.base.ContentVisualBase;

import javafx.geometry.Bounds;
import javafx.geometry.Dimension2D;
import javafx.geometry.Rectangle2D;
import javafx.geometry.Side;
import javafx.scene.Node;
import javafx.scene.paint.Color;
import javafx.scene.shape.Rectangle;

/** Base class for {@link ContentVisualBase content visuals} depicted by rectangles. */
public abstract class RectangularContentVisualBase extends ContentVisualBase {
	/** Constructor. */
	public RectangularContentVisualBase(IContentMVCBundle mvcb) {
		super(mvcb, new Rectangle(), new Rectangle());
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);

		Rectangle2D bounds = getCurrentBounds();
		double leftX = bounds.getMinX();
		double upperY = bounds.getMinY();
		double width = bounds.getWidth();
		double height = bounds.getHeight();
		if(enableVisual()) {
			Rectangle rectangle = getVisualRectangleShape();
			rectangle.setX(leftX);
			rectangle.setY(upperY);
			rectangle.setWidth(width);
			rectangle.setHeight(height);
			Dimension2D cornerArc = getCornerArcDimensions();
			rectangle.setArcWidth(cornerArc.getWidth());
			rectangle.setArcHeight(cornerArc.getHeight());
			rectangle.setOpacity(getOpacity());
			rectangle.setStroke(getBorderColor());
			rectangle.setStrokeWidth(getBorderWidth());
			rectangle.setStrokeType(getBorderType());
			rectangle.setFill(getFillColor());
		}

		if(enableInteraction()) {
			Rectangle ha = getHitAreaRectangleShape();
			double lnk = getHitAreaStartLinkSize();
			double resize = getHitAreaResizeSize();
			ha.setX(leftX - lnk - resize / 2);
			ha.setY(upperY - lnk - resize / 2);
			ha.setWidth(lnk + resize + width + lnk);
			ha.setHeight(lnk + resize + height + lnk);
			Color interactionColor = getInteractionColor();
			ha.setStroke(interactionColor);
			ha.setFill(interactionColor);
		}
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getAnchorageLocation(IContentAnchorageVisual visual) {
		Rectangle2D pb = getCurrentBounds();
		Side side = visual.getLayout(ISideLayout.class).getSide();
		double offset = visual.getLayout(IOffsetLayout.class).getOffset();
		RectangularBorderLocation rbl =
				new RectangularBorderLocation(side, offset, pb, visual.getDimensions());
		return rbl.getLocation();
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getLinkAnchorage(DiagramCoordinate indication) {
		Rectangle2D b = getCurrentBounds();
		Dimension2D dummy = new Dimension2D(1, 1);
		RectangularBorderLocation loc = getClosestLocationOnBounds(
				indication.add(-b.getMinX(), -b.getMinY()), b, dummy, dummy);
		return loc.getLocation().add(b.getMinX(), b.getMinY());
	}

	/** Returns the dimensions of the corner arcs. */
	protected Dimension2D getCornerArcDimensions() {
		return new Dimension2D(15, 15);
	}

	/** {@inheritDoc} */
	@Override
	protected EDragGesture dragGestureHitTest(Node node, DiagramCoordinate diagramLocation) {
		if(node == getHitAreaShape()) {
			double x = diagramLocation.getLocalX(node);
			double y = diagramLocation.getLocalY(node);
			double l = getHitAreaStartLinkSize();
			double r = getHitAreaResizeSize();
			Bounds bounds = node.getBoundsInLocal();
			// check for move area
			double inset = l + r;
			Rectangle2D moveArea = new Rectangle2D(inset, inset, bounds.getWidth() - 2 * inset,
					bounds.getHeight() - 2 * inset);
			if(moveArea.contains(x, y)) {
				return EDragGesture.MOVE;
			}
			if(x < l || y < l || x > bounds.getWidth() - l || y > bounds.getHeight() - l) {
				return EDragGesture.NEW_LINK;
			}
			if(x > bounds.getWidth() - inset && y >= l && y <= bounds.getHeight() - inset) {
				return EDragGesture.RESIZE_H;
			}
			if(y > bounds.getHeight() - inset && x >= l && x <= bounds.getWidth() - inset) {
				return EDragGesture.RESIZE_V;
			}
			if(x > bounds.getWidth() - inset && y > bounds.getHeight() - inset) {
				return EDragGesture.RESIZE_VH;
			}
			return EDragGesture.NONE;
		}
		if(node == getVisualShape() || node == text) {
			return EDragGesture.MOVE;
		}
		return EDragGesture.NONE;
	}

	/** Returns the visual rectangle shape. */
	protected final Rectangle getVisualRectangleShape() {
		// wild cast works: see constructor
		return (Rectangle)getVisualShape();
	}

	/** Returns the hit area rectangle shape. */
	protected final Rectangle getHitAreaRectangleShape() {
		// wild cast works: see constructor
		return (Rectangle)getHitAreaShape();
	}

	/**
	 * Returns the extension of the hit area in pixel used to trigger link creation. This number
	 * specifies how many pixels beyond the visible border are used to trigger link creation by a
	 * mouse drag gesture.
	 */
	protected double getHitAreaStartLinkSize() {
		return getViewer().getFeatures().getMaximumSpacing();
	}

	/** Returns the size of the hit area in pixel used to resize the content visual. */
	protected double getHitAreaResizeSize() {
		return getViewer().getFeatures().getMaximumSpacing();
	}
}
