/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.mesh;

import static javafx.scene.paint.Color.WHITE;
import static javafx.scene.shape.VertexFormat.POINT_NORMAL_TEXCOORD;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import javafx.scene.Group;
import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.MeshView;
import javafx.scene.shape.TriangleMesh;

/** Class representing a mesh. */
public class Mesh3D {
	/** The mesh object name. */
	private final String name;
	/** The list of components of this mesh. */
	private final List<MeshComponent> components;
	/** The list of vertices of this mesh. */
	private final List<Vertex3D> vertices;
	/** The list of normals of this mesh. */
	private final List<Normal3D> normals;
	/** The list of texture coordinates of this mesh. */
	private final List<Texture2D> textures;
	/** The JavaFX group of this mesh. */
	private Group javaFXGroup = null;

	/** Constructor. */
	public Mesh3D(String name, List<Vertex3D> vertices, List<Texture2D> textures,
			List<Normal3D> normals, List<MeshComponent> meshComps) {
		this.name = name;
		this.vertices = vertices;
		this.normals = normals;
		// ensure texture list not null (in case no textures are used)
		this.textures = textures == null ? new LinkedList<>() : textures;
		this.components = new ArrayList<>(meshComps);

		// ensure at least one element is present in the texture list
		if(this.textures.isEmpty()) {
			this.textures.add(new Texture2D(0, 0));
		}
	}

	/** Adds the given material and faces as a new component of this mesh. */
	public void add(List<Face3D> faces, Material3D material, boolean smoothShading) {
		components.add(new MeshComponent(material, faces, smoothShading));
	}

	/** Returns the name. */
	public String getName() {
		return name;
	}

	/** Returns the components. */
	public List<MeshComponent> getComponents() {
		return components;
	}

	/** Class for mesh components. */
	protected static final class MeshComponent {
		/** The material of this component. */
		private Material3D material;
		/** The faces of this component. */
		private List<Face3D> faces;
		/** Smooth shading flag. */
		private boolean smoothShading;

		/** Constructor. */
		public MeshComponent(Material3D material, List<Face3D> faces, boolean smoothShading) {
			this.material = material;
			this.faces = new ArrayList<>(faces);
			this.smoothShading = smoothShading;
		}

		/** Returns the material. */
		public Material3D getMaterial() {
			return material;
		}

		/** Sets the material. */
		public void setMaterial(Material3D material) {
			this.material = material;
		}

		/** Returns the faces. */
		public List<Face3D> getFaces() {
			return faces;
		}

		/** Returns the smooth shading flag. */
		public boolean getSmoothShading() {
			return smoothShading;
		}

		/** Sets the smooth shading flag. */
		public void setSmoothShading(boolean smoothShading) {
			this.smoothShading = smoothShading;
		}

		/** Returns the {@link MeshView} object for this mesh component. */
		public MeshView getJavaFXShape(Mesh3D mesh) {
			int load = faces.size() * 3 / 2;
			HashMap<Integer, Integer> vIndexMap = new HashMap<>(load);
			HashMap<Integer, Integer> tIndexMap = new HashMap<>(load);
			HashMap<Integer, Integer> nIndexMap = new HashMap<>(load);
			TriangleMesh m = new TriangleMesh(POINT_NORMAL_TEXCOORD);
			for(Face3D f : faces) {
				addVertex(f.getVertex0(), mesh, m, vIndexMap, tIndexMap, nIndexMap);
				addVertex(f.getVertex1(), mesh, m, vIndexMap, tIndexMap, nIndexMap);
				addVertex(f.getVertex2(), mesh, m, vIndexMap, tIndexMap, nIndexMap);
			}
			MeshView mv = new MeshView(m);
			Color diff = material != null ? material.getDiffuse() : WHITE;
			PhongMaterial pm = new PhongMaterial(diff);
			mv.setMaterial(pm);
			return mv;
		}

		/** Adds the given vertex to the JavaFX mesh. */
		private void addVertex(IndexPointer vertex, Mesh3D objMesh, TriangleMesh tMesh,
				HashMap<Integer, Integer> vIndexMap, HashMap<Integer, Integer> tIndexMap,
				HashMap<Integer, Integer> nIndexMap) {
			Integer vIndexInTM = vIndexMap.get(vertex.getVertexIndex());
			if(vIndexInTM == null) {
				Vertex3D v = vertex.getVertex(objMesh);
				tMesh.getPoints().addAll((float)v.getX(), (float)v.getY(), (float)v.getZ());
				vIndexInTM = vIndexMap.size();
				vIndexMap.put(vertex.getVertexIndex(), vIndexInTM);
			}

			Integer tIndexInTM = tIndexMap.get(vertex.getTextureIndex());
			if(tIndexInTM == null) {
				Texture2D t = vertex.getTexture(objMesh);
				tMesh.getTexCoords().addAll((float)t.getU(), (float)t.getV());
				tIndexInTM = tIndexMap.size();
				tIndexMap.put(vertex.getTextureIndex(), tIndexInTM);
			}

			Integer nIndexInTM = nIndexMap.get(vertex.getNormalIndex());
			if(nIndexInTM == null) {
				Normal3D n = vertex.getNormal(objMesh);
				tMesh.getNormals().addAll((float)n.getX(), (float)n.getY(), (float)n.getZ());
				nIndexInTM = nIndexMap.size();
				nIndexMap.put(vertex.getNormalIndex(), nIndexInTM);
			}
			tMesh.getFaces().addAll(vIndexInTM, nIndexInTM, tIndexInTM);
		}
	}

	/** Returns the JavaFX group representing this mesh. */
	public Group toJavaFXShape() {
		if(javaFXGroup == null) {
			javaFXGroup = new Group();
			for(MeshComponent mc : components) {
				javaFXGroup.getChildren().add(mc.getJavaFXShape(this));
			}
		}
		return javaFXGroup;
	}

	/**
	 * Pointer to elements in indexed lists representing all the data in a vertex of a
	 * {@link Face3D}.
	 */
	public static class IndexPointer {
		/** The vertex position index. */
		private final int vertexIndex;
		/** The texture coordinate index. */
		private final int textureIndex;
		/** The normals index. */
		private final int normalIndex;

		/** Constructor. */
		public IndexPointer(int v, int t, int n) {
			this.vertexIndex = v;
			this.textureIndex = t;
			this.normalIndex = n;
		}

		/** Constructor. */
		public IndexPointer(int v, int n) {
			// texture list always contains at least one element
			this(v, 0, n);
		}

		/** Returns the vertex index. */
		public int getVertexIndex() {
			return vertexIndex;
		}

		/** Returns the texture index. */
		public int getTextureIndex() {
			return textureIndex;
		}

		/** Returns the normal index. */
		public int getNormalIndex() {
			return normalIndex;
		}

		/** Returns the vertex. */
		public Vertex3D getVertex(Mesh3D mesh) {
			return mesh.vertices.get(vertexIndex);
		}

		/** Returns the texture. */
		public Texture2D getTexture(Mesh3D mesh) {
			return mesh.textures.get(textureIndex);
		}

		/** Returns the normal. */
		public Normal3D getNormal(Mesh3D mesh) {
			return mesh.normals.get(normalIndex);
		}
	}
}
