/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 the ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.pref;

import org.conqat.ide.commons.ui.swt.SWTUtils;
import org.conqat.lib.commons.collections.PairList;
import org.conqat.lib.commons.string.StringUtils;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.preference.ListEditor;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

/**
 * A field editor for managing a list of string mappings, which is represented
 * as a {@link PairList}.
 * <p>
 * A clean implementation would use a {@link TableViewer} with two columns to
 * display the mappings. However, as the effort of implementing a full field
 * editor is significant and there are no suitable base classes, we use an
 * implementation is based on {@link ListEditor} as a compromise.
 * 
 * @author $Author: $
 * @version $Rev: $
 * @ConQAT.Rating GREEN Hash: 40205EE6D2478FCB79CB6606A9B8D399
 */
public class MappingListEditor extends ListEditor {

	/**
	 * This separator is used when storing a mapping as a single string and is
	 * also visible to the user. Thus, the separator should not be too ugly and
	 * also not be likely to be contained in any path.
	 */
	private static final String PAIR_SEPARATOR = " -> ";

	/**
	 * This separator is used to concatenate the individual mappings. Is is not
	 * visible to the user but also should not contain characters that have a
	 * special meaning in a regex.
	 */
	private static final String ELEMENT_SEPARATOR = "#<=>#";

	/**
	 * The list control we hooked the {@link #listDoubleClickListener} to. This
	 * is required, because of the way the base class manages the underlying
	 * list. Also see {@link #getListControl(Composite)} for details.
	 */
	private List hookedListControl = null;

	/** Double click listener that implements editing. */
	private final SelectionListener listDoubleClickListener = new SelectionAdapter() {
		/** {@inheritDoc} */
		@Override
		public void widgetDefaultSelected(SelectionEvent e) {
			int index = hookedListControl.getSelectionIndex();
			if (index < 0) {
				return;
			}
			String newValue = queryMappingPair(hookedListControl.getItem(index));
			if (newValue != null) {
				hookedListControl.setItem(index, newValue);
				selectionChanged();
			}
		}
	};

	/** The title of the dialog that is opened. */
	private final String dialogTitle;

	/** The label for the mapping's "from" field. */
	private final String fromTitle;

	/** The label for the mapping's "to" field. */
	private final String toTitle;

	/** Constructor. */
	public MappingListEditor(String name, String labelText, Composite parent,
			String dialogTitle, String fromTitle, String toTitle) {
		super(name, labelText, parent);
		this.dialogTitle = dialogTitle;
		this.fromTitle = fromTitle;
		this.toTitle = toTitle;

		// if the getListControl() method was called in super(), we could not
		// hook our listener as it was still null at this point. Thus, we hook
		// it now by calling getListControl()
		getListControl(parent);
	}

	/** {@inheritDoc} */
	@Override
	protected String createList(String[] items) {
		return StringUtils.concat(items, ELEMENT_SEPARATOR);
	}

	/** {@inheritDoc} */
	@Override
	protected String getNewInputObject() {
		// using PAIR_SEPARATOR essentially indicates an empty mapping
		return queryMappingPair(PAIR_SEPARATOR);
	}

	/** {@inheritDoc} */
	@Override
	protected String[] parseString(String stringList) {
		String[] parts = stringList.split(ELEMENT_SEPARATOR);
		if (parts.length == 1 && StringUtils.isEmpty(parts[0])) {
			return new String[0];
		}
		return parts;
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * The problem with the base class is that is can potentially recreate a
	 * control, as the class is reused which the UI is not. Additionally, this
	 * method is not a pure factory method but rather uses lazy initialization.
	 */
	@Override
	public List getListControl(Composite parent) {
		List listControl = super.getListControl(parent);

		// also check whether listDoubleClickListener is null, as this method
		// can be called from the super constructor, when the field will not be
		// initialized yet.
		if (hookedListControl == listControl || listDoubleClickListener == null) {
			return listControl;
		}

		if (hookedListControl != null) {
			hookedListControl.removeSelectionListener(listDoubleClickListener);
		}

		hookedListControl = listControl;

		if (hookedListControl != null) {
			hookedListControl.addSelectionListener(listDoubleClickListener);
		}

		return listControl;
	}

	/** Queries the user for a mapping pair. Returns null on cancel. */
	private String queryMappingPair(String initialMappingValue) {
		MappingInputDialog dialog = new MappingInputDialog(initialMappingValue,
				getShell());
		if (dialog.open() != Window.OK) {
			return null;
		}
		return dialog.getMapping();
	}

	/** Joins a mapping pair into a single string. */
	private static String joinMapping(String from, String to) {
		return from + PAIR_SEPARATOR + to;
	}

	/** Splits a mapping string into an array of size 2. */
	private static String[] splitMapping(String mapping) {
		String[] pair = mapping.split(PAIR_SEPARATOR, 2);
		if (pair.length == 1) {
			return new String[] { pair[0], StringUtils.EMPTY_STRING };
		}
		return pair;
	}

	/**
	 * Converts a concatenated preference value to a list of mappings
	 * represented as a {@link PairList}.
	 */
	public static PairList<String, String> deserializeMappings(
			String concatenatedValue) {
		PairList<String, String> result = new PairList<String, String>();
		for (String mapping : concatenatedValue.split(ELEMENT_SEPARATOR)) {
			if (!StringUtils.isEmpty(mapping)) {
				String[] pair = splitMapping(mapping);
				result.add(pair[0], pair[1]);
			}
		}
		return result;
	}

	/** Serializes mappings given as a {@link PairList} into a single string. */
	public static String serializeMappings(PairList<String, String> mappings) {
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < mappings.size(); ++i) {
			if (i > 0) {
				sb.append(ELEMENT_SEPARATOR);
			}
			sb.append(joinMapping(mappings.getFirst(i), mappings.getSecond(i)));
		}
		return sb.toString();
	}

	/** A dialog for entering a mapping. */
	private class MappingInputDialog extends Dialog {

		/** The initial from value. */
		private String from;

		/** The initial to value. */
		private String to;

		/** Text field for the "from" part. */
		private Text fromText;

		/** Text field for the "to" part. */
		private Text toText;

		/** Constructor. */
		public MappingInputDialog(String initialValuePair, Shell shell) {
			super(shell);
			String[] pair = splitMapping(initialValuePair);
			from = pair[0];
			to = pair[1];
		}

		/** {@inheritDoc} */
		@Override
		protected void configureShell(Shell newShell) {
			super.configureShell(newShell);
			newShell.setText(dialogTitle);
		}

		/** {@inheritDoc} */
		@Override
		protected Control createDialogArea(Composite parent) {
			Composite composite = (Composite) super.createDialogArea(parent);
			((GridLayout) composite.getLayout()).numColumns = 2;

			SWTUtils.createLabel(composite, fromTitle);
			fromText = SWTUtils.createText(composite, from, null);
			SWTUtils.createLabel(composite, toTitle);
			toText = SWTUtils.createText(composite, to, null);

			return composite;
		}

		/** {@inheritDoc} */
		@Override
		protected void okPressed() {
			// store strings before disposal
			from = fromText.getText();
			to = toText.getText();
			super.okPressed();
		}

		/** Returns the current mapping. */
		public String getMapping() {
			return joinMapping(from, to);
		}
	}
}
