/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 The ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.xmlmodel;

import java.util.ArrayList;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Element;

import org.conqat.lib.commons.collections.IdentityHashSet;
import org.conqat.lib.commons.xml.XMLUtils;

/**
 * Cache for model elements that wrap XML elements. The purpose of this cache is
 * to ensure that the same model element instances are returned as long as the
 * underlying XML elements are the same. (In other words, the goal of the cache
 * is not to increase performance.)
 * <p>
 * This cache automatically maintains its state to remove unused model element
 * instances to prevent loitering model element instances from causing memory
 * leaks.
 * 
 * @param <T>
 *            Type of model element that gets cached
 * 
 * @author hummelb
 * @author doelle
 * @author $Author: deissenb $
 * @version $Rev: 34252 $
 * @levd.rating GREEN Hash: 04EF82B8B49599651B7CBA18F097C12B
 */
public abstract class ElementCache<T> {

	/** Maps from XML elements to their corresponding model elements */
	private final Map<Element, T> cache = new IdentityHashMap<Element, T>();

	/**
	 * Retrieves the corresponding model elements for the the XML elements. If a
	 * model element for a XML element has been retrieved before, the same
	 * instance is returned, else a new instance of the model element is
	 * created.
	 * <p>
	 * This method performs cache maintenance and removes model elements from
	 * the cache that were not retrieved during the last call in order to avoid
	 * memory leaks due to loitering objects.
	 */
	public List<T> getModelElementsFor(List<Element> elements) {
		List<T> result = new ArrayList<T>();
		Set<Element> notNeeded = new IdentityHashSet<Element>(cache.keySet());
		for (Element e : elements) {
			result.add(obtainModelElement(e));
			notNeeded.remove(e);
		}

		cache.keySet().removeAll(notNeeded);
		return result;
	}

	/**
	 * Checks if the cache contains a wrapping model element for a single XML
	 * element, creates it if necessary and returns it.
	 */
	public T obtainModelElement(Element e) {
		T parameter = cache.get(e);
		if (parameter == null) {
			parameter = createFromXmlElement(e);
			cache.put(e, parameter);
		}
		return parameter;
	}

	/**
	 * Template method that deriving classes override to create their specific
	 * model elements
	 */
	protected abstract T createFromXmlElement(Element e);

	/**
	 * Returns a {@link List} of model elements for a given xmlTag in a DOM
	 * node.
	 * 
	 * @param <T>
	 *            the result lists type
	 * @param domNode
	 *            the DOM node
	 * @param xmlTag
	 *            the XML tag to be searched for
	 * @param cache
	 *            the cache mapping from XML elements to their corresponding
	 *            model elements
	 * @return a list of model elements
	 */
	public static <T> List<T> getModelElements(Element domNode, String xmlTag,
			ElementCache<T> cache) {
		return cache.getModelElementsFor(XMLUtils.getNamedChildren(domNode,
				xmlTag));
	}

}